import numpy as np
import random


class LabelledDataGenerator(object):
    def __init__(self, label, window_size, prefetch_size, loaded_mat_files, num_features, stride, seed, shuffle=True, flatten = True ):
        self.label = label
        self.window_size = window_size
        self.prefetch_size = prefetch_size
        self.num_features = num_features
        self.loaded_mat_files = loaded_mat_files
        self.stride = stride
        self.mat_idx = -1
        self.sample_size = self.num_features*self.window_size
        self.shuffle = shuffle
        np.random.seed(seed)
        # TODO (shaurya) : consider contiguous numpy array instead of python list
        self.data_buffer = list()
        self.flatten = flatten
        # self.data_buffer = np.empty((self.prefetch_size, self.sample_size), dtype=np.float32)
        # self.data_buffer_idx = 0
        # self.data_buffer_length = 0

        self._fill_buffer()


    def _fill_buffer(self):
        while len(self.data_buffer) < self.prefetch_size:
            self.mat_idx += 1
            self.mat_idx %= len(self.loaded_mat_files)

            mat_data = self.loaded_mat_files[self.mat_idx]
            label_sequences = getattr(mat_data['label_sequences'], self.label)


            if len(label_sequences) == 0:
                continue
            elif len(label_sequences.shape) == 1:
                label_sequences = np.expand_dims(label_sequences, axis=1)

            features = mat_data['X']
            assert(features.shape[0] == self.num_features)

            for label_sequence in label_sequences.T:

                start = label_sequence[0]
                end = label_sequence[1]

                feature_sequence = features[:, start : end]
                if self.flatten:
                    feature_sequence_flattened = np.reshape( feature_sequence, (-1,), order='F' ).astype(np.float32)

                    stride = self.stride*self.num_features
                    flattened_window_size = self.window_size*self.num_features
                    for i in range(0, len(feature_sequence_flattened), stride):
                        if i+flattened_window_size > len(feature_sequence_flattened):
                            break

                        self.data_buffer.append(feature_sequence_flattened[i:i+flattened_window_size])
                else:
                    for i in range(0,feature_sequence.shape[1],self.stride):
                        if i+self.window_size > feature_sequence.shape[1]:
                            break

                        self.data_buffer.append(feature_sequence[:,i:i+self.window_size])


        if self.shuffle:
            np.random.shuffle(self.data_buffer)

    def get_batch(self, batch_size):
        generator = self.get_sample()
        while True:
            if len(self.data_buffer) < batch_size:
                self._fill_buffer()

            batch_data = np.empty((batch_size, self.sample_size))
            for i in range(batch_size):
                sample = next(generator)
                batch_data[i] = sample

            yield batch_data


    def get_sample(self):
        while True:
            if not self.data_buffer:
                self._fill_buffer()

            yield self.data_buffer.pop(0)



