import scipy.io as sio
import glob
from os import listdir, mkdir
from os.path import join, isdir, isfile, basename
import numpy as np
import random
import scipy
from ExternalStudents.keras_utils.data_io import ThreadSafeIterator
from ExternalStudents.keras_utils.data_io.ThreadSafeIterator import ThreadSafeGenerator
from ExternalStudents.keras_utils.data_io.LabelledDataGenerator import LabelledDataGenerator
from enum import Enum


class LabellingMethod(Enum):
    CENTER = 1
    MOST_FREQUENT = 2
    END_POINT = 3
    FN_END_POINT = 4



_FAR_LABEL = 'Far'
_NEAR_LABEL = 'Near'
_FAR2NEAR_LABEL = 'Far2Near'
_NEAR2FAR_LABEL = 'Near2Far'

class MatDataGenerator(object):
    def __init__(self, mat_folder,
                 window_size,
                 seed,
                 prefetch_size,
                 class_labels,
                 batch_size=32,
                 debug_mode=False,
                 debug_output_folder="C:\\temp\EirikLid\debug",
                 shuffle=True,
                 class_weight=None,
                 data_stride=1,
                 flatten=True,
                 mat_files=None):

        self.num_classes = len(class_labels)

        assert(batch_size%self.num_classes==0)
        assert(data_stride > 0)

        if debug_mode:
            if debug_output_folder is None:
                raise ValueError('Debug mode enabled, but debug output folder is None')
            elif not isdir(debug_output_folder):
                print('debug output folder {} does not exist. Creating.'.format(debug_output_folder))
                mkdir(debug_output_folder)


        self.debug_mode = debug_mode
        self.debug_output_folder = debug_output_folder
        self.stride = data_stride
        self.mat_folder = mat_folder
        self.class_labels = class_labels
        self.batch_size = batch_size


        self.window_size = window_size
        self.flat_window = flatten
        self.seed = seed

        np.random.seed(seed)

        ######### CROSS VAL FILES VS NONE #################
        if mat_files == None:
            mat_pattern = join(mat_folder, "*.mat")
            mat_files = glob.glob(mat_pattern)
            if len(mat_files) == 0:
                raise ValueError('Mat file folder {} does not contain any mat files'.format(mat_folder))
        ######### CROSS VAL FILES VS NONE #################

        if shuffle:
            np.random.shuffle(mat_files)

        self.loaded_files = [sio.loadmat(mf, struct_as_record=False, squeeze_me=True) for mf in mat_files]

        for mat_data in self.loaded_files:
            self._assign_label_sequences(mat_data, LabellingMethod.END_POINT)

        self.class_weight = np.ones((self.num_classes,), dtype=np.float32)/self.num_classes
        if class_weight != None:
            self.class_weight = class_weight/np.sum(class_weight)


        self.num_features = getattr(self.loaded_files[0]['metadata'], 'features_length')
        if self.flat_window:
            self.input_shape = self.num_features*self.window_size
        else:
            self.input_shape = (self.num_features,self.window_size)

        if self.debug_mode:
            self.debug_data_buffer = np.empty((100*batch_size, self.input_shape), dtype=np.float32)
            self.debug_label_buffer = np.empty((100*batch_size, 1), dtype=np.float32)
            self.debug_example_idx = 0
            self.debug_file_idx = 1


        self.labelled_data_generators = dict()
        for label in self.class_labels:
            label_data_gen = LabelledDataGenerator(label, window_size, prefetch_size, \
                                                   self.loaded_files, self.num_features, self.stride, \
                                                   seed, shuffle, flatten = flatten)

            label_data_generator = label_data_gen.get_sample()
            self.labelled_data_generators[label] = label_data_generator


    def _assign_label_sequences(self, mat_data, labelling_method):
        y = mat_data['y']
        mat_data['label_sequences'] = scipy.io.matlab.mio5_params.mat_struct()
        zero = np.where(y == 0)[0]
        zero_split = np.split(zero, np.where(np.diff(zero) != 1)[0] + 1)

        near = np.where(y!=0)[0]
        near_split = np.split(near, np.where(np.diff(near) != 1)[0]+1)

        far2near = np.where(y == 2)[0]
        far2near_split = np.split(far2near, np.where(np.diff(far2near) != 1)[0] + 1)

        near2far = np.where(y == 3)[0]
        near2far_split = np.split(near2far, np.where(np.diff(near2far) != 1)[0] + 1)

        if len(near) == 0:
            setattr(mat_data['label_sequences'], _FAR_LABEL, np.array([[0], [len(y)-1]], dtype=np.int32))
            setattr(mat_data['label_sequences'], _NEAR_LABEL, np.array([], dtype=np.int32))
            setattr(mat_data['label_sequences'], _FAR2NEAR_LABEL, np.array([], dtype=np.int32))
            setattr(mat_data['label_sequences'], _NEAR2FAR_LABEL, np.array([], dtype=np.int32))
            return


        near_start_idx = near[0]
        near_end_idx = near[-1]

        # Assumes only Far - Near - Far recordings
        if labelling_method == LabellingMethod.MOST_FREQUENT:
            far_seq_start = [0, near_start_idx - (self.window_size // 2) - 1]
            far_seq_end = [near_end_idx - (self.window_size // 2) + 1, len(y) - 1]
            near_seq = [near_start_idx - (self.window_size // 2), near_end_idx + (self.window_size // 2)]
            setattr(mat_data['label_sequences'], _FAR_LABEL, np.array([[far_seq_start[0], far_seq_end[0]],
                                                                       [far_seq_start[1], far_seq_end[1]]],
                                                                      dtype=np.int32))
            setattr(mat_data['label_sequences'], _NEAR_LABEL, np.array([[near_seq[0]], [near_seq[1]]], dtype=np.int32))

            # far_mat = [[ far_seq_start[0], far_seq_end[0] ], [ far_seq_start[1], far_seq_end[1] ] ]
            # near_mat = [ [near_seq[0]],[near_seq[1]] ]
            # setattr(mat_data['label_sequences'], _FAR_LABEL, np.array(far_mat, dtype=np.int32))
            # setattr(mat_data['label_sequences'], _NEAR_LABEL, np.array(near_mat, dtype=np.int32))

        elif labelling_method == LabellingMethod.END_POINT:
            far_mat = self.seq_split_to_mat(zero_split)
            near_mat = self.seq_split_to_mat(near_split)
            far2near_mat = self.seq_split_to_mat(far2near_split)
            near2far_mat = self.seq_split_to_mat(near2far_split)
            setattr(mat_data['label_sequences'], _FAR_LABEL, np.array(far_mat, dtype=np.int32))
            setattr(mat_data['label_sequences'], _NEAR_LABEL, np.array(near_mat, dtype=np.int32))
            setattr(mat_data['label_sequences'], _FAR2NEAR_LABEL, np.array(far2near_mat, dtype=np.int32))
            setattr(mat_data['label_sequences'], _NEAR2FAR_LABEL, np.array(near2far_mat, dtype=np.int32))


        else:
            raise NotImplementedError('Labelling method implemented')


    def seq_split_to_mat(self,split):
        if split[0].size:
            start = []
            end = []

            for seq in split:
                start.append(max(0, seq[0] - (self.window_size)))
                end.append(seq[-1] - (self.window_size))
            return [start, end]
        else:
            return []


    def _label_idx_to_key(self, idx):
        return self.class_labels[idx]

    def input_shape(self):
        return self.input_shape

    def class_weight(self):
        return self.class_weight

    def _dump_debug_data(self, batch_data, batch_labels, batch_size):
        ii = self.debug_example_idx
        self.debug_data_buffer[ii : ii + batch_size] = batch_data
        self.debug_label_buffer[ii : ii + batch_size, 0] = np.argmax(batch_labels, axis=1)

        if (self.debug_example_idx + batch_size) == self.debug_data_buffer.shape[0]:
            file_path = join(self.debug_output_folder, 'batch_samples_' + str(self.debug_file_idx) + '.mat')
            data_dict = {'samples' : self.debug_data_buffer, 'labels' : self.debug_label_buffer}
            sio.savemat(file_path, data_dict)
            self.debug_file_idx += 1

        self.debug_example_idx = (self.debug_example_idx + batch_size) % self.debug_data_buffer.shape[0]

    @ThreadSafeGenerator
    def generator(self):
        batch_size = self.batch_size

        while True:
            if self.flat_window:
                batch_data = np.empty((batch_size, self.num_features * self.window_size), dtype=np.float32) # Used on flattened
            else:
                batch_data = np.empty((batch_size, self.num_features, self.window_size), dtype=np.float32)

            batch_labels = np.zeros((batch_size, self.num_classes), dtype=np.float32)

            for i in range(batch_size):
                label_idx = np.random.choice(self.num_classes, p=self.class_weight)
                label = self._label_idx_to_key(label_idx)
                batch_data[i] = next(self.labelled_data_generators[label])
                batch_labels[i, label_idx] = 1.0

            if self.debug_mode:
                self._dump_debug_data(batch_data, batch_labels, batch_size)

            yield batch_data, batch_labels



#window_size = 40
#mat_folder = r'C:\temp\PixelIMU\20180312\matfiles'
#seed = 0
#prefetch_size = 1024

#train_datagen_config = {
#       'window_size' : window_size,
#       'mat_folder' : mat_folder,
#       'seed' : seed,
#       'prefetch_size' : prefetch_size,
#       'class_labels' : [_FAR_LABEL, _NEAR_LABEL]
#   }

#mtdg = MatDataGenerator(**train_datagen_config)

#for i, (bd, bl) in enumerate(mtdg.generator()):
#    if i == 2000:
#        break