import scipy.io as sio
import glob
from os import listdir, mkdir
from os.path import join, isdir, isfile, basename
import numpy as np
import random
import scipy
from keras_utils.data_io import ThreadSafeIterator
from keras_utils.data_io.ThreadSafeIterator import ThreadSafeGenerator
from keras_utils.data_io.LabelledDataGenerator import LabelledDataGenerator
from enum import Enum


class LabellingMethod(Enum):
    CENTER = 1
    MOST_FREQUENT = 2
    END_POINT = 3



_FAR_LABEL = 'Far'
_NEAR_LABEL = 'Near'
_FAR2NEAR_LABEL = 'Far2Near'
_NEAR2FAR_LABEL = 'Near2Far'

class SequentialDataGenerator(object):
    def __init__(self, mat_folder,
                 window_size,
                 seed,
                 prefetch_size,
                 class_labels,
                 batch_size=32,
                 debug_mode=False,
                 debug_output_folder="C:\\temp\EirikLid\debug",
                 shuffle=True,
                 class_weight=None,
                 data_stride=1,
                 flatten=True,
                 mat_files=None):

        self.num_classes = len(class_labels)

        assert(batch_size%self.num_classes==0)
        assert(data_stride > 0)

        if debug_mode:
            if debug_output_folder is None:
                raise ValueError('Debug mode enabled, but debug output folder is None')
            elif not isdir(debug_output_folder):
                print('debug output folder {} does not exist. Creating.'.format(debug_output_folder))
                mkdir(debug_output_folder)


        self.debug_mode = debug_mode
        self.debug_output_folder = debug_output_folder
        self.stride = data_stride
        self.mat_folder = mat_folder
        self.class_labels = class_labels
        self.batch_size = batch_size


        self.window_size = window_size
        self.flatten = flatten
        self.seed = seed

        np.random.seed(seed)

        ########## SEQ ########
        self.mat_idx = -1
        self.data_buffer = list()
        self.label_buffer = list()
        self.prefetch_size = prefetch_size

        ########## SEQ ########


        ######### CROSS VAL FILES VS NONE #################
        if mat_files == None:
            mat_pattern = join(mat_folder, "*.mat")
            mat_files = glob.glob(mat_pattern)
            if len(mat_files) == 0:
                raise ValueError('Mat file folder {} does not contain any mat files'.format(mat_folder))
        ######### CROSS VAL FILES VS NONE #################

        if shuffle:
            np.random.shuffle(mat_files)

        self.loaded_files = [sio.loadmat(mf, struct_as_record=False, squeeze_me=True) for mf in mat_files]

        self.num_features = getattr(self.loaded_files[0]['metadata'], 'features_length')
        if self.flatten:
            self.input_shape = self.num_features*self.window_size # Used with flatten
        else:
            self.input_shape = (self.num_features,self.window_size)

        if self.debug_mode:
            self.debug_data_buffer = np.empty((100*batch_size, self.input_shape), dtype=np.float32)
            self.debug_label_buffer = np.empty((100*batch_size, 1), dtype=np.float32)
            self.debug_example_idx = 0
            self.debug_file_idx = 1




    def _fill_buffer(self):
        while len(self.data_buffer) < self.prefetch_size:

            #### ToDo: Dette skal inn i end_of_timeseries ####
            self.mat_idx += 1
            self.mat_idx %= len(self.loaded_files)
            mat_data = self.loaded_files[self.mat_idx]


            features = mat_data['X']
            labels = mat_data['y']
            assert (features.shape[0] == self.num_features)


            if self.flatten:
                feature_sequence_flattened = np.reshape(features, (-1,), order='F').astype(np.float32)

                stride = self.stride * self.num_features
                flattened_window_size = self.window_size * self.num_features
                for i in range(0, len(feature_sequence_flattened), stride):
                    if i + flattened_window_size > len(feature_sequence_flattened):
                        break

                    self.data_buffer.append(feature_sequence_flattened[i:i + flattened_window_size])
                    self.label_buffer.append(labels[i//self.num_features])
            else:
                for i in range(0, features.shape[1], self.stride):
                    if i + self.window_size > features.shape[1]:
                        break

                    self.data_buffer.append(features[:, i:i + self.window_size])
                    self.label_buffer.append(labels[i+self.window_size-1])


    def input_shape(self):
        return self.input_shape

    def class_weight(self):
        return self.class_weight

    def _dump_debug_data(self, batch_data, batch_labels, batch_size):
        ii = self.debug_example_idx
        self.debug_data_buffer[ii : ii + batch_size] = batch_data
        self.debug_label_buffer[ii : ii + batch_size, 0] = np.argmax(batch_labels, axis=1)

        if (self.debug_example_idx + batch_size) == self.debug_data_buffer.shape[0]:
            file_path = join(self.debug_output_folder, 'batch_samples_' + str(self.debug_file_idx) + '.mat')
            data_dict = {'samples' : self.debug_data_buffer, 'labels' : self.debug_label_buffer}
            sio.savemat(file_path, data_dict)
            self.debug_file_idx += 1

        self.debug_example_idx = (self.debug_example_idx + batch_size) % self.debug_data_buffer.shape[0]

    def get_batch(self, batch_size):
        generator = self.get_sample()
        while True:
            if len(self.data_buffer) < batch_size:
                self._fill_buffer()

            batch_data = np.empty((batch_size, self.num_features*self.window_size))
            batch_labels = np.empty((batch_size, self.num_classes))
            for i in range(batch_size):
                sample = next(generator)
                batch_data[i] = sample[0]
                batch_labels[i] = sample[1]

            yield batch_data, batch_labels


    def get_sample(self):
        while True:
            if not self.data_buffer:
                self._fill_buffer()

            yield (self.data_buffer.pop(0), self.label_buffer.pop(0)) #ToDo: consider using popleft()

    @ThreadSafeGenerator
    def generator(self):
        batch_size = self.batch_size
        generator = self.get_sample()
        while True:
            if len(self.data_buffer) < batch_size:
                self._fill_buffer()
            if self.flatten:
                batch_data = np.empty((batch_size, self.num_features*self.window_size))
            else:
                batch_data = np.empty((batch_size, self.num_features, self.window_size))
            batch_labels = np.zeros((batch_size, self.num_classes))
            for i in range(batch_size):
                sample = next(generator)
                batch_data[i] = sample[0]
                batch_labels[i][sample[1]] = 1.0

            if self.debug_mode:
                self._dump_debug_data(batch_data, batch_labels, batch_size)

            yield batch_data, batch_labels



#window_size = 40
#mat_folder = r'C:\temp\PixelIMU\20180312\matfiles'
#seed = 0
#prefetch_size = 1024

#train_datagen_config = {
#       'window_size' : window_size,
#       'mat_folder' : mat_folder,
#       'seed' : seed,
#       'prefetch_size' : prefetch_size,
#       'class_labels' : [_FAR_LABEL, _NEAR_LABEL]
#   }

#mtdg = MatDataGenerator(**train_datagen_config)

#for i, (bd, bl) in enumerate(mtdg.generator()):
#    if i == 2000:
#        break


