from tensorflow.python.tools import freeze_graph
from tensorflow.core.framework import graph_pb2
from tensorflow.python.platform import gfile
from tensorflow.python.tools import optimize_for_inference_lib
from tensorflow.python.framework import dtypes
from keras import backend as K
from os.path import join as path_join
from os.path import splitext
from os.path import isdir
from os import mkdir
import tensorflow as tf
from keras.models import load_model, Sequential, clone_model
from keras.layers import Dropout, GRU
import tempfile


from tensorflow.python.framework import graph_util
from tensorflow.python.framework import graph_io

# Must be called within the context of a Keras+tensorflow session

def _remove_dropout(model, input_node_name, output_node_name):
    model_result = Sequential()
    for k in model.layers:

        if type(k) is Dropout:
            continue

        model_result.add(k)
    assert(len(model.layers) >= 2)
    model_result.layers[0].name = input_node_name
    # model_result.layers[-1].name = output_node_name
    return model_result

def _has_GRU(model):
    for layer in model.layers:
        if type(layer) is GRU:
            return True
    return False


def save_optimized_graph(
    model,
    logdir,
    graph_file_name,
    input_node_name,
    output_node_name,
    output_file_path='',
    variable_names_blacklist='',
    remove_dropout=True):

    if not isdir(logdir):
        print('Creating folder {}'.format(logdir))
        mkdir(logdir)

    if '.pb' != graph_file_name[-3:]:
        raise ValueError("Graph file name {} must have .pb file extension".format(graph_file_name))
    graph_root_name = splitext(graph_file_name)[0]

    model.save(path_join(logdir, graph_root_name + '_training' + '.h5'))
    # Save Checkpoint


    K.set_learning_phase(0)
    config = model.get_config()
    weights = model.get_weights()

    if remove_dropout:
        new_model = Sequential.from_config(config)
        new_model.set_weights(weights)
        inference_model = _remove_dropout(new_model, input_node_name, output_node_name)
        keras_model_file = path_join(logdir, graph_root_name + '_inference' + '.h5')
        inference_model.save(keras_model_file)
        inference_model = load_model(keras_model_file)

    sess = K.get_session()
    saver = tf.train.Saver(tf.global_variables())


    # Save GraphDef
    tf.train.write_graph(sess.graph_def, logdir, graph_file_name)
    checkpoint = path_join(logdir, graph_root_name)
    saver.save(sess, checkpoint, write_meta_graph=True)

    # Protobuf saving modifies names of nodes
    pb_graph_input_name = input_node_name
    pb_graph_output_name = output_node_name


    frozen_graph_name = graph_root_name + '_frozen.pb'
    frozen_graph_file = path_join(logdir, frozen_graph_name)

    # Freeze graph with weights
    freeze_graph.freeze_graph(input_graph=path_join(logdir,graph_file_name),
             input_saver="",
             input_checkpoint=checkpoint,
             output_graph=frozen_graph_file,
             input_binary=False,
             output_node_names=output_node_name,
             restore_op_name="save/restore_all",
             filename_tensor_name="save/Const:0",
             clear_devices=True,
             initializer_nodes="",
             variable_names_blacklist=variable_names_blacklist)

    if not _has_GRU(model):
        if output_file_path is '':
            output_file_path = graph_root_name + '_optimized.pb'

        input_graph_def = graph_pb2.GraphDef()
        optimized_graph_path = path_join(logdir, output_file_path)

        pb_graph_input_name = input_node_name
        with gfile.Open(frozen_graph_file, "rb") as f:
            frozen_graph_data = f.read()

        input_graph_def.ParseFromString(frozen_graph_data)

        output_graph_def = optimize_for_inference_lib.optimize_for_inference(
            input_graph_def,
            pb_graph_input_name.split(","),
            pb_graph_output_name.split(","),
            dtypes.float32.as_datatype_enum)

        with gfile.FastGFile(optimized_graph_path, "wb") as f:
            f.write(output_graph_def.SerializeToString())

        input_dim = model.input_shape[1]
        print("Saved optimized graph to : {}".format(optimized_graph_path))
        print('\n\nTFlite toco conversion command :')
        print('/home/lamer/Development/tensorflow/tools/toco --input_file=/home/lamer/temp/{0}_optimized.pb --output_file=/home/lamer/temp/{0}.lite --input_format=TENSORFLOW_GRAPHDEF --output_format=TFLITE --inference_type=FLOAT --input_shape={1},1 --input_array={2} --output_array={3}'.format(graph_root_name, input_dim, pb_graph_input_name, pb_graph_output_name))

    else:
        print('model has GRU layer, skipping optimize for inference ')



def save_optimized_graph_simple(
    model_dir,
    logdir,
    graph_file_name,
    input_node_name,
    output_node_name,
    output_file_path='',
    variable_names_blacklist='',
    remove_dropout=True):

    if not isdir(logdir):
        print('Creating folder {}'.format(logdir))
        mkdir(logdir)

    if '.pb' != graph_file_name[-3:]:
        raise ValueError("Graph file name {} must have .pb file extension".format(graph_file_name))
    graph_root_name = splitext(graph_file_name)[0]

    K.set_learning_phase(0)
    # K.set_image_data_format('channels_last')
    net_model = load_model(model_dir)

    sess = K.get_session()
    constant_graph = graph_util.convert_variables_to_constants(sess, sess.graph.as_graph_def(), [output_node_name])
    graph_io.write_graph(constant_graph, logdir, graph_file_name, as_text=False)

